#include <QTimer>
#include <QApplication>
#include <QLabel>
#include <QImage>

/*
 * Applies a 3x3 convolution filter and returns the resulting image
 */
QImage sobel(const QImage &orig){
  QImage img = orig;
  int w = img.width();
  int h = img.height();
  static double sobelX[3][3] = {{-1.0, 0.0, 1.0}, 
                                {-2.0, 0.0, 2.0}, 
				{-1.0, 0.0, 1.0}}; // horiz Sobel edge detect
  static double sobelY[3][3] = {{1.0, 2.0, 1.0}, 
                                {0.0, 0.0, 0.0}, 
				{-1.0, -2.0, -1.0}}; // vert Sobel edge detect
  // convolution filter  
  for(int y=0;y<h;y++)
    for(int x=0;x<w;x++){
       qreal totx=0.0;
       qreal toty=0.0;
       
       for(int sy = 0; sy<3;sy++)
         for(int sx=0; sx<3;sx++){
           int tx = x-1+sx;
           int ty = y-1+sy;
           if(tx>=0 && tx<w && ty>=0 && ty<h){
             int gray = qGray(orig.pixel(tx, ty));
             totx+=gray*sobelX[sx][sy];
             toty+=gray*sobelY[sy][sx];
           }
         }
      int tot = qBound(0, qAbs((int)totx)+qAbs((int)toty), 255);
      img.setPixel(x, y, qRgb(tot, tot, tot));
    }
  return img;
}

/*
 *  Computes path with least energy
 */
QVector<int> findPath(const QImage &orig){
  int w = orig.width();
  int h = orig.height();
  
  qreal **map = new qreal*[w];
  for(int i=0;i<w;i++){
    map[i] = new qreal[h];
    for(int j=0;j<h;j++) map[i][j] = 2000000;
  }
  for(int i=0;i<w;i++) map[i][0] = 0.0;
  for(int y=1;y<h;y++)
      for(int x=0;x<w;x++){
        // assume down is best
        int best = x;
        qreal bestCost = map[x][y-1];
        // check left and right
        for(int c=x-1;c<=x+1;c+=2){
          if(c>=0 && c<w && map[c][y-1]<bestCost){
            best = c;
            bestCost = map[c][y-1];
          }
        }
        map[x][y] = map[best][y-1] + qRed(orig.pixel(x,y));
    }
    qreal bestCost = map[0][h-1];
    int nextX = 0;
    QVector<int> path(h);
    // find lowest energy
    QVector<int> bests;
    bests << nextX;
    for(int i=1;i<w;i++){
      if(map[i][h-1]==bestCost){
        bests << nextX;
      } else if(map[i][h-1]<=bestCost){
        nextX = i;
        bestCost = map[i][h-1];
        bests.clear();
        bests << nextX;
      }
    }
    // this is different than the original
    // chooses a random path from ones with the lowest energy
    // instead of choosing the one most to the left 
    if(bests.size()>1) nextX = bests.at(qrand() % bests.size());
    
    // build path
    path[h-1] = nextX;
    for(int y=h-2;y>=0;y--){
      int tryX = nextX;
      double costX = map[nextX][y];
      for(int tx = nextX-1; tx<=nextX+1; tx+=2){
        if(tx>=0 && tx<w && map[tx][y]<costX){
          costX=map[tx][y];
          tryX=tx;
        }
      }
      nextX = tryX;
      path[y] = tryX;
    }
    
    // dispose of map  
    for(int i=0;i<w;i++) delete [] map[i];
    delete [] map;
    return path;
}

/*
 * draws the path on the image
 */
QImage applyPath(QImage img, const QVector<int> &path){
  int h = img.height();
  for(int i=0;i<h;i++){
    img.setPixel(path[i], i, qRgb(0, 255, 0));
  }
  return img;
}

/*
 * blends three pixels together
 */
QRgb blend3(QRgb l, QRgb c, QRgb r){
  int _r = qRound((qreal)(qRed(l)+qRed(c)+qRed(r))/3.0);
  int _g = qRound((qreal)(qGreen(l)+qGreen(c)+qGreen(r))/3.0);  
  int _b = qRound((qreal)(qBlue(l)+qBlue(c)+qBlue(r))/3.0);
  int _a = qRound((qreal)(qAlpha(l)+qAlpha(c)+qAlpha(r))/3.0);
  return qRgba(_r, _g, _b, _a);
}

/*
 * blends two pixels together
 */
QRgb blend2(QRgb l, QRgb r){
  int _r = qRound((qreal)(qRed(l)+qRed(r))/2.0);
  int _g = qRound((qreal)(qGreen(l)+qGreen(r))/2.0);  
  int _b = qRound((qreal)(qBlue(l)+qBlue(r))/2.0);
  int _a = qRound((qreal)(qAlpha(l)+qAlpha(r))/2.0);
  return qRgba(_r, _g, _b, _a);
}

/*
 * reduces the image using the path
 *
 * uses memcpy to quickly copy image data. 
 * Works only for true color images
 */
QImage reduce(const QImage &orig, const QVector<int> &path){
  QImage img(orig.width()-1, orig.height(), orig.format());
  int h = orig.height();
  int w = orig.width();
  for(int y=0;y<h;y++){
    int rx = path[y];
    const QRgb *sourceline = (const QRgb*)orig.scanLine(y);
    QRgb *destline = (QRgb*)img.scanLine(y);
    if(rx>0){
      memcpy(destline, sourceline, rx*sizeof(QRgb));
      if(rx<w-1){
        destline[rx-1] = blend2(sourceline[rx-1], sourceline[rx]  );
        destline[rx]   = blend2(sourceline[rx],   sourceline[rx+1]);
        memcpy(destline+rx+1, sourceline+2+rx, (w-2-rx)*sizeof(QRgb));
      }
    } else memcpy(destline, sourceline+1, (w-1)*sizeof(QRgb));
  }
  return img;
}

/*
 * Performs one full iteration of the algorithm
 */
QImage narrower(const QImage &orig){
  QImage s = sobel(orig);
  QVector<int> path = findPath(s);
  return reduce(orig, path);
}


class Reducer : public QLabel {
  Q_OBJECT
public:
  Reducer(const QImage &img, QWidget *parent=0) 
  : QLabel(parent), image(img){
    destWidth = img.width();
    setAlignment(Qt::AlignCenter);
    setImage(image);
  }
  void shrink(int dw, int dh){
    destWidth = dw;
    destHeight = dh;
    lastWidth = false;
    next();
  }
private:
  QVector<int> path;
  QImage image;
  int destWidth;
  int destHeight;
  bool lastWidth;
  void setImage(const QImage &img){
    setPixmap(QPixmap::fromImage(img));
  }
private slots:
  void next(){
    if(image.width()>destWidth || image.height()>destHeight){
      if(lastWidth && image.height()==destHeight) lastWidth = false;
      if(!lastWidth && image.width()==destWidth) lastWidth = true;
      calculatePath();
    } else setImage(image);
  }
  void calculatePath(){
      QImage im = image;
      if(lastWidth){
        QMatrix m;
        m = m.rotate(90);
        im = image.transformed(m);
      }
      QImage s = sobel(im);
      path = findPath(s);
      QImage newim = applyPath(im, path);
      if(lastWidth){
        QMatrix m;
        m = m.rotate(-90);
        newim = newim.transformed(m);
      } else {
          setImage(newim);
      }
      QTimer::singleShot(0, this, SLOT(reduceImage()));
  }
  void reduceImage(){
      if(lastWidth){
        QMatrix m;
        m = m.rotate(90);
        image = image.transformed(m);
      }
      image = reduce(image, path);
      if(lastWidth){
        QMatrix m;
        m = m.rotate(-90);
        image = image.transformed(m);
      }
      setImage(image);
      lastWidth = !lastWidth;
      QTimer::singleShot(0, this, SLOT(next()));
  }
};

#include "main.moc"

int main(int argc, char **argv){
  QApplication app(argc, argv);
  const QStringList& args = app.arguments();
  if(args.size()<2)
    qFatal("Usage: %s image [newwidth [newheight]]", 
           qPrintable(args.at(0)));
  QImage img(args.at(1));
  if(img.width()>800)
    img = img.scaled(800, 600, Qt::KeepAspectRatio, Qt::SmoothTransformation);
  Reducer lab(img);
  lab.shrink(args.size()>2 ? args.at(2).toInt() : qRound(img.width()*.9), 
             args.size()>3 ? args.at(3).toInt() : qRound(img.height()*.9)
  );
  lab.show();
  return app.exec();
}
